from abc import abstractmethod

from meiga import NotImplementedMethodError, Result

from petisco.base.misc.interface import Interface


class AppService(Interface):
    @abstractmethod
    def execute(self, *args, **kwargs) -> Result:
        return NotImplementedMethodError



import json
from typing import Dict, Optional, Type, Union

from petisco.base.domain.message.message import Message


class Command(Message):
    def __init__(self, **kwargs):
        self._set_data()
        self._set_attributes(**kwargs)
        self.type = "command"

    @staticmethod
    def from_dict(message_data: Dict, target_type: Optional[Type] = None):
        target_type = Command if target_type is None else target_type
        data = message_data.get("data")
        command = target_type()
        command._set_data(**data)
        return command

    @staticmethod
    def from_json(message_json: Union[str, bytes], target_type: Optional[Type] = None):
        event_dict = json.loads(message_json)
        return Command.from_dict(event_dict, target_type)

    def __repr__(self):
        return self.to_str(class_name="Command", type="domain_event")



from abc import ABC, abstractmethod
from types import FunctionType
from typing import Any, List

from meiga import Error, NotImplementedMethodError, Result

from petisco.base.misc.result_mapper import ResultMapper, default_failure_handler
from petisco.base.misc.wrapper import wrapper


def get_mapper(bases, config):
    mapper = ResultMapper()
    if config:
        for base in bases:
            method = getattr(base, "get_config_mapper", None)
            if method:
                mapper = method(config)

    else:
        for base in bases:
            method = getattr(base, "get_default_mapper", None)
            if method:
                mapper = method()

    return mapper


class MetaController(type, ABC):
    middlewares: List = {}

    def __new__(mcs, name, bases, namespace):
        config = namespace.get("Config")

        mapper = get_mapper(bases, config)

        if "execute" not in namespace:
            raise NotImplementedError(
                "Petisco Controller must implement an execute method"
            )

        new_namespace = {}
        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                attribute = wrapper(attribute, name, config, mapper)
            new_namespace[attributeName] = attribute

        return super().__new__(mcs, name, bases, new_namespace)

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError


class Controller(metaclass=MetaController):
    @staticmethod
    def get_default_mapper() -> ResultMapper:
        return ResultMapper()

    @staticmethod
    def get_config_mapper(config):
        return ResultMapper(
            error_map=getattr(config, "error_map", None),
            success_handler=getattr(config, "success_handler", lambda result: result),
            failure_handler=default_failure_handler,
        )

    @abstractmethod
    def execute(self, *args, **kwargs) -> Result[Any, Error]:
        return NotImplementedMethodError



from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage


class NotImplementedNotifier(Notifier):
    def publish(self, message: NotifierMessage):
        pass

    def publish_exception(self, notifier_exception_message: NotifierExceptionMessage):
        pass



from abc import abstractmethod

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message_bus import MessageBus


class DomainEventBus(MessageBus):
    @abstractmethod
    def publish(self, domain_event: DomainEvent):
        raise NotImplementedError

    @abstractmethod
    def retry_publish_only_on_store_queue(self, domain_event: DomainEvent):
        raise NotImplementedError

    def _check_is_domain_event(self, domain_event: DomainEvent):
        if not domain_event or not issubclass(domain_event.__class__, DomainEvent):
            raise TypeError("DomainEventBus only publishes DomainEvent objects")

    @abstractmethod
    def close(self):
        raise NotImplementedError



from petisco.extra.elastic.is_elastic_available import is_elastic_available

rabbitmq = list()
if is_elastic_available():

    from petisco.extra.elastic.elastic_apm_monitoring_app_service import (
        ElasticApmMonitoringAppService,
    )
    from petisco.extra.elastic.elastic_connection import ElasticConnection
    from petisco.extra.elastic.elastic_database import ElasticDatabase

    elastic = ["ElasticConnection", "ElasticDatabase", "ElasticApmMonitoringAppService"]
else:
    elastic = []

__all__ = elastic



def assert_http(response, expected_status_code: int):
    response_detail = (
        response.json if isinstance(response.json, dict) else response.json()
    )
    assert (
        response.status_code == expected_status_code
    ), f"{response.status_code} - {response_detail}"



from abc import abstractmethod
from typing import List, Type

from meiga import BoolResult

from petisco.base.domain.message.message import Message
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class AllMessageSubscriber(MessageSubscriber):
    def subscribed_to(self) -> List[Type[Message]]:
        return [Message]

    @abstractmethod
    def handle(self, message: Message) -> BoolResult:
        raise NotImplementedError()



from abc import ABC, abstractmethod


class ControllerExecutorWrapper(ABC):
    @abstractmethod
    def execute(self, api_controller, *args, **kwargs):
        pass


class DefaultControllerExecutorWrapper(ControllerExecutorWrapper):
    def execute(self, api_controller, *args, **kwargs):
        api_controller._execute_controller(*args, **kwargs)  # noqa



from abc import abstractmethod
from typing import List, Type

from meiga import BoolResult

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class DomainEventSubscriber(MessageSubscriber):
    @abstractmethod
    def subscribed_to(self) -> List[Type[DomainEvent]]:
        raise NotImplementedError()

    @abstractmethod
    def handle(self, domain_event: DomainEvent) -> BoolResult:
        raise NotImplementedError()



from typing import List

from petisco.base.domain.message.message_configurer import MessageConfigurer
from petisco.base.domain.message.message_subscriber import MessageSubscriber


class NotImplementedMessageConfigurer(MessageConfigurer):
    def configure_subscribers(self, subscribers: List[MessageSubscriber]):
        pass

    def clear(self):
        pass



from typing import Optional

from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.model.uuid import Uuid


class AlreadyExists(DomainError):
    ...


class ClientAlreadyExists(DomainError):
    ...


class UserAlreadyExists(DomainError):
    ...


class AggregateAlreadyExistError(DomainError):
    def __init__(
        self,
        aggregate_id: Uuid,
        repository_name: Optional[str] = None,
        table_name: Optional[str] = None,
    ):
        aggregate_id_str = f" ({aggregate_id.value})" if aggregate_id else ""
        repository_str = f" (repository: {repository_name})" if repository_name else ""
        table_str = f" (table: {table_name})" if repository_name else ""
        message = (
            f"Aggregate{aggregate_id_str} already exist{repository_str}{table_str}"
        )
        uuid_value = aggregate_id.value if aggregate_id else None
        super().__init__(uuid_value=uuid_value, additional_info={"message": message})
